@@ -8,6 +8,17 @@ from einops import rearrange
 def exists(val):
     return val is not None
 
+# decorators
+
+def eval_decorator(fn):
+    def inner(self, *args, **kwargs):
+        was_training = self.training
+        self.eval()
+        out = fn(self, *args, **kwargs)
+        self.train(was_training)
+        return out
+    return inner
+
 # tensor helpers
 
 def log(t, eps = 1e-20):
